/* ************************************************************

   Program:   Simple (Sequential Imputation for Multi-Point 
	      Linkage 
	      Estimation)
   File:      simple.c
   Author:    Mark E. Irwin
   Date:      July 27, 1999
   Purpose:   Program to do multi-locus linkage analysis by 
	      sequential imputation methods. 
   Modified:  Nov 21, 2001 use cM instead of M
   Modified:  1/25/02 incorporate complex peeler.  The original 
	      version is saved as simple.c

************************************************************* */

#include <stdio.h>   /* fprint */
#include <string.h>  /* strcpy */
#include "simple.h"  /* aff */

#define  FIRSTMODEL     0
#define  to_cmorgan      0
#define  to_recomb_frac 1
#define  PID            0
#define  MAX_NAME       20
#define  TRUE           1
#define  FALSE          0

int main(int argc, char *argv[])
{
  int   i,j,k,model,locus,iter; /* for indexing only */
  
  int     nmarkers;
  bool    recomb_fracs, lod=FALSE; /* boolean */
  FILE    *runfile, *gfopen();

/* ReadLinkageFile stuff */
  double  **liability;
  int     no_liab_class;
  int     *nalleles;
  int     nloci;
  double  *mdistances;
  int     *markerorder;
  int     isdisease;
  int     nQT;
  double  disaf[2];
  double  **maf;

/* ReadPedFile */
  int    npedigrees;
  int    *pedigreesize;
  int    ***pedigree;
  aff    **affected;
  int    ****marker;
  double ****qpheno;
  char   **pednames;

/* Commands with their default values */
  cmd     c;
  char    command[MAX_NAME];


  int    ***ancestry;

/* for simple case */
  int    **peelorder;
  int    **peeltype;
  int    **peelinto;

/* for complex case */
  int    *npeels;
  sindex **cutset, **peelset;
  int    **ntrans, **nrinto;
  tindex  ***tidx; 
  Rindex  ***ridx;

  int    ****simhap;
  int    ***shap;
  int    *hap;
  double **distances;
  int    **locusorder;
  double ***likdistances;
  int    *numlik;

  int    ***spacesize;
  int    *****haplabel;
  int    ****ibdobs;
  int    ****ibdsim;
  int    *ibdF;
  int    *ibdM;
  int    **jpotflag;
  int    ***blank;
  double ****ppotential;
  double ****wppotential;
  double *****jpotential;
  double ****simwppotential;
  double *****simjpotential;
  double **prob;
  double ***likprob;
  double **likratio;
  double *weight;
  double ibdobsprob;

/* for complex case */
  int    ****cutspaces, ***cutspace, ****peelspaces, ***peelspace;
  int    ***nualleles,****ualleles;
  double ****rfunction, *simprob, ****wpp;
  int    *whaps, *whaps2;

  double CalcWeight();
  double ComplexCalcWeight();
  double GetProbIBDObs();
  double CalcLikRatio();

  void ReadLinkageFile();
  void ReadPedFile();
  void ParseCommands();
  void GetAncestry();
  void ReadRecombFracs();
  void ReadDistances();
  void CreateDistances();
  void CalcPeelingOrder();
  void ReadPeelFiles();
  void CalcBlank();
  void CreateSpaces();
  void ComplexCreateSpaces();
  void ComplexGetHaplabel();
  void GetHaplabel();
  void GetIBDObs();
  void IBDBlank();
  void IBDReset();
  void MapVector();             /* Zak added */
  void MapMatrix();		/* Zak added */
  void CHS();			/* Zak added */
  void Calcibdprobs();
  void Calcmarginprobs();
  void CreateMarkerPPots();
  void CreateDiseasePrePPots();
  void CreateDiseasePPots();
  void GetWPP();
  void GetTransIndex();
  void GetRIndex();
  void CopyIBD();
  void Peel();
  void ComplexPeel();
  void Simulate();
  void ComplexSimulate();
  void UpdateIBDsim();
  void DumpCommands();
  void DumpIBDS();
  void DumpHaps();
  void SimBlank();

  void DebugReadRun();
  void DebugReadPedigree();
  void DebugReadRecombFracs();
  void DebugCreateDistances();
  void DebugReadDistances();
  void DebugGetAncestry();
  void DebugCalcPeelingOrder();
  void DebugReadPeelFiles();
  void DebugReadMarker();
  void DebugCalcBlank();
  void DebugCreateSpaces();
  void DebugComplexCreateSpaces();
  void DebugGetHaplabel();
  void DebugGetIBDObs();
  void DebugCreateDistances();
  void DebugMapMatrix();
  void DebugCalcmarginprobs();
  void DebugProbtable();
  void DebugGetProbIBDObs();
  void DebugCreateMarkerPPots();
  void DebugReadDiseasePrePPots(); 
  void DebugGetWPP();
  void DebugGetTransIndex();
  void DebugGetRIndex();
  void DebugPeel();
  void DebugComplexPeel();
  void DebugSimulate();
  void DebugComplexSimulate();

  void srand48();

  if(argc>2) nerror("USAGE: simple <commandfile (optional)>");

/* allocate space */
  c.load.par= (char *) calloc(MAX_NAME, sizeof(char));
  c.scan.par= (char *) calloc(MAX_NAME, sizeof(char));

/* default LociName, PedName */
  strcpy(c.load.par,"linkloci.dat");
  strcpy(c.scan.par,"ped.dat");

/* find out if there is a load or scan command.  If so then get
   the filenames, otherwise use the above defaults */ 


  if(argc==2){
    runfile = gfopen(argv[1],"r");
    while(fscanf(runfile,"%s",command)==1){
      if(strcmp(command,"load")==0){
        if(fscanf(runfile,"%s",c.load.par)!=1)
          nerror("expecting a file name after command 'load'");}
      else if(strcmp(command,"scan")==0){
        if(fscanf(runfile,"%s",c.scan.par)!=1)
          nerror("expecting a file name after command 'scan'");}
      else if(strcmp(command,"missingQT")==0){
        if(fscanf(runfile,"%lf",&c.missingQT.par)!=1)
          nerror("expecting a real number after command 'missingQT'");
        else c.missingQT.def=FALSE;
      }
    }
  }

  ReadLinkageFile(&nloci,
		  &mdistances,
		  &markerorder,
		  &nalleles,
		  &maf,
		  &isdisease,
		  &nQT,
		  disaf,
		  &liability,
		  &no_liab_class,
		  c.load.par);

  /* nloci = nloci - nQT; */
  nmarkers = nloci - isdisease;


  ReadPedFile(&npedigrees,
  	      &pednames,	
	      &pedigreesize,
	      &pedigree,
	      &marker,
              nmarkers,
	      nQT,
	      &qpheno,
	      isdisease,
	      &affected,
	      no_liab_class,
	      c); 

  ParseCommands(&c,
		argc,
		runfile,
		&nloci,
		nQT,
		nalleles,
		npedigrees,
		isdisease);

  if(c.debug.par) DebugReadRun(c.debug.par,
                               c.interference.model,
                               c.interference.par,
			       nmarkers,
                               &c.file,
			       mdistances,
			       c.nsteps.par,
			       isdisease,
			       c.processorder.par,
			       c.maxiter.par,
			       c.seed.par);


  if(c.debug.par) DebugReadPedigree(npedigrees,
			            pedigreesize,
			            pedigree);

  if(c.debug.par) DebugReadMarker(npedigrees,
                                  pedigreesize,
                                  pedigree,
			          nmarkers,
                                  nalleles,
                                  marker,
			          maf,
				  nQT,
				  qpheno);

  srand48(c.seed.par);

  /*nloci = nmarkers+isdisease;*/
/* I can't do this inside ParseCommands because things get messy 
with ignore_disease.
1/27/01 I made ignore_disease obsolete */
  if((strcmp(c.analysis.par,"LOD")==0) ||(strcmp(c.analysis.par,"BOTH")==0)) lod=TRUE;
  c.file.nmodels=1;
  if(lod && c.file.def) /* if we are to create the distances */
  {
    if(nmarkers==1) c.file.nmodels=2;
    else c.file.nmodels=nmarkers;
  }
  /* if(nmarkers==1 && !lod && c.file.def) c.file.nmodels=0; */

  /* nmodels=1;this was in the mlla.c .  I'm not sure why. 
  It was probably because Mark just wanted to dump ibds and didn't
  want to calculate lod scores */
 
  /* if nloci=1, ie no disease and one marker then we have to
  allocate weight is just a scalar */
  /* if(nloci==1)
    weight = (double *) calloc(1,sizeof(double));
  else */
    weight = (double *) calloc(nloci+c.file.nmodels-1-!lod,sizeof(double)); 

  GetAncestry(npedigrees,
	      pedigreesize,
	      pedigree,
	      &ancestry);
  if(c.debug.par) DebugGetAncestry(npedigrees,
			           pedigreesize,
			           pedigree,
			           ancestry); 

/* if there are no peel files to read */
  if(!c.peel.par){
    CalcPeelingOrder(npedigrees,
                     pedigreesize,
		     ancestry,
                     &peelorder,
                     &peeltype,
                     &peelinto);
    if(c.debug.par) DebugCalcPeelingOrder(npedigrees,
                                          pedigreesize, 
				          pedigree,
                                          peelorder,
                                          peeltype,
                                          peelinto);

    CreateSpaces(npedigrees,
                 pedigreesize,
	         ancestry,
                 nalleles,
                 nmarkers,
                 marker,
                 lod,
	         nloci,
	         peeltype,
	         peelinto,
                 &spacesize,
                 &ibdobs,
                 &ibdsim,
	         &ibdF,
	         &ibdM,
	         &jpotflag,
	         &ppotential,
	         &wppotential,
	         &jpotential,
	         &simwppotential,
	         &simjpotential,
	         &simhap,
	         &shap,
	         &hap);
    if(c.debug.par) DebugCreateSpaces(npedigrees,
                                      pedigreesize,
                                      pedigree,
                                      nmarkers,
                                      marker,
                                      spacesize,
			              jpotflag);

    GetHaplabel(npedigrees,
	        pedigreesize,
	        ancestry,
	        nalleles,
	        marker,
	        nmarkers,
	        lod,
	        nloci,
	        spacesize,
	        &haplabel);


    if(lod)  
    {
      CreateDiseasePrePPots(ppotential,
			    affected,  
			    liability,
			    nloci,
			    npedigrees,
			    pedigreesize,
			    pedigree);
      if(c.debug.par) DebugReadDiseasePrePPots(npedigrees,
		                               pedigreesize,
			 	               pedigree,
			 	               nloci,
				               disaf,
			 	               ppotential);

    } 
  /* create equivalent for complex case later */
    CalcBlank(npedigrees,
              pedigreesize,
	      ancestry,
              nmarkers,
	      nloci,
	      marker,
	      ppotential,
              &blank);
    if(c.debug.par) DebugCalcBlank(npedigrees,
			           pedigreesize,
			           pedigree,
			           nloci,
			           blank);
  }
  else { /* read the peel files */
    ReadPeelFiles(c.peel.fname,
                  npedigrees,
                  pedigreesize,
                  &npeels,
                  &cutset, 
                  &peelset);
    if(c.debug.par) DebugReadPeelFiles(npedigrees,
                                       npeels,
                                       cutset,
                                       peelset);


    GetCutPeelPedIndex(npedigrees,
                       pedigree,
                       pedigreesize,
                       npeels,
                       cutset,
                       peelset);
    if(c.debug.par) DebugGetCutPeelPedIndex(npedigrees,
                                            npeels,
                                            cutset,
                                            peelset);


    ComplexCreateSpaces(npedigrees,
                        pedigreesize,
                        npeels,
                        cutset,
                        peelset,
                        ancestry,
                        nalleles,
                        nmarkers,
                        marker,
                        nloci,
                        &ibdobs,
                        &ibdsim,
                        &ibdF,
                        &ibdM,
                        &simhap,
                        &shap,
                        &hap,
			&spacesize,
                        &cutspaces,
                        &cutspace,
                        &peelspaces,
                        &peelspace,
			&wpp,
			&rfunction,
                        &simprob,
                        &whaps,
                        &whaps2,
			&nualleles,
			&ualleles);
    if(c.debug.par) DebugComplexCreateSpaces(npedigrees,
                                             npeels,
                                             nmarkers,
                                             marker,
                                             cutset,
                                             peelset,
                                             cutspaces,
                                             cutspace,
                                             peelspaces,
                                             peelspace);

    ComplexGetHaplabel(npedigrees,
	        pedigreesize,
	        ancestry,
	        nalleles,
	        marker,
	        nmarkers,
	        lod,
	        nloci,
	        spacesize,
	        nualleles,
	        ualleles,
	        &haplabel);
  }


  if(c.debug.par) DebugGetHaplabel(npedigrees,
			           pedigreesize,
			           nloci,
			           spacesize,
			           haplabel);

  if(c.file.nmodels>1) IBDBlank(npedigrees,
			        pedigreesize,
			        c.processorder.par[nloci-1]-1,
			        ibdobs);

  GetIBDObs(npedigrees,
            pedigreesize,
            ancestry,
            nmarkers,
            nalleles,
            marker,
	    lod,
	    nloci,
            ibdobs);
  if(c.debug.par) DebugGetIBDObs(npedigrees,
                                 pedigreesize,
                                 pedigree,
                                 nmarkers,
                                 ibdobs); 
     
/* If units are not specified then we determine if we have 
 cM or recomb fracs using the preplink
 criteria that if at least one > .5 we have cM */

  if(c.units.def){
    recomb_fracs = TRUE;
    strcpy(c.units.par,"rec-frac");
    for(i=0; i < nmarkers-1; i++)
      if(mdistances[i] > 0.5) {
        recomb_fracs = FALSE;
        strcpy(c.units.par,"cM");
      }
  }
  else{
    if(strcmp(c.units.par,"cM")==0) recomb_fracs=FALSE;
    else if(strcmp(c.units.par,"rec-frac")==0) recomb_fracs=TRUE;
  }


  /* the default is to create distances vs reading them */
  /* if we are not doing lod scores then we ignore the file.  This file must */
  /* be intended for scan and not simple */
  /* the CreateDistances function is set up to handle the */
  /* case when we just want to simulate in-phase genoyptes for NPL analysis */
  if(!c.file.def && lod) 
    if(recomb_fracs==TRUE){
    /* this won't work right unless the file has the same number
       of models as specified by c.file.nmodels */
      ReadRecombFracs(nloci,
		      c.file.nmodels,
		      &locusorder,
		      &distances, /* actually recomb fracs */
		      &numlik,
		      &likdistances);
      if(c.debug.par) DebugReadRecombFracs(nloci,
      				           c.file.nmodels,
				           locusorder,
		                           distances,
				           numlik,
				           likdistances);
    }
    else{
    /* this won't work right unless the file has the same number
       of models as specified by c.file.nmodels */
      ReadDistances(nloci,
                    &locusorder,
                    &distances, 
                    &numlik,
                    &likdistances,
                    c.file);
      if(c.debug.par) DebugReadDistances(nloci,
                                         c.file.nmodels,
                     	                 lod,
                                         locusorder,
                                         distances,
                                         numlik,
                                         likdistances);
    }
  
  else {
  /* simulate the disease in the middle of each adjacent marker */
  /* where the measure is in terms of genetic distance NOT */
  /* recombination fractions */
    if(recomb_fracs==TRUE)
      MapVector(nmarkers-1, mdistances, to_cmorgan); 
    CreateDistances(nloci,
                    c.file.nmodels,
                    lod,
                    c.nsteps.par,
                    mdistances,
                    &locusorder,
                    &distances,
                    &numlik,
                    &likdistances);
    if(c.debug.par) DebugCreateDistances(nloci,
                                         c.file.nmodels,
                       		         lod,
                                         locusorder,
                                         distances,
                                         numlik,
                                         likdistances);
  
  }
  switch(c.interference.model){
    case 0:
    /* use recomb fracs assuming no interference */ 
      if(recomb_fracs==FALSE){
      /* map cmorgans to recomb fracs */
        MapMatrix(nloci,
                  c.file.nmodels,
                  distances,
                  numlik,
                  likdistances,
	          lod,
                  to_recomb_frac);
        if(c.debug.par) DebugMapMatrix(nloci,
                                       c.file.nmodels,
                                       distances,
                                       numlik,
                                       likdistances,
			               lod,
                                       to_recomb_frac);
      }
      Calcibdprobs(nloci,
	           c.file.nmodels,
	           locusorder,
	           distances,  /* actually transformed to rec fracs */
	           &prob,
	           numlik,
	           likdistances, /* actually transformed to rec fracs */
	           &likprob);
      break; 
    case 19:
      CHS(nloci,
          c.file.nmodels,
	  lod,
          locusorder,
          distances,
          numlik,
          likdistances,
          &prob,
          &likprob,
          c.interference.par);
      break;
    default: 
      nerror("There is no proper interference model!");
  }

  Calcmarginprobs(nloci,
	  	  c.file.nmodels,
		  prob,
		  numlik,
		  likprob);
  if(c.debug.par) DebugCalcmarginprobs(nloci,
				       c.file.nmodels,
                                       prob,
				       numlik,
				       likprob);

  if(c.debug.par) DebugProbtable(nloci,
			         c.file.nmodels,
			         prob,
			         numlik,
			         likprob);

  likratio = (double **) calloc(c.file.nmodels,sizeof(double *));
  for(k=0; k<c.file.nmodels; k++)
    likratio[k] = (double *) calloc(numlik[k],sizeof(double));

  ibdobsprob = GetProbIBDObs(npedigrees,
			     pedigreesize,
			     ibdobs,
			     ibdF,
			     ibdM,
			     nloci,
			     FIRSTMODEL,
			     prob);

  if(c.debug.par) DebugGetProbIBDObs(npedigrees,
                                     pedigreesize,
			             pedigree,
                                     ibdobs,
                                     ibdF,
                                     ibdM,
                                     nloci,
			             FIRSTMODEL,
                                     prob);

  if(!c.peel.par) { /* not complex */
    CreateMarkerPPots(npedigrees,
		      pedigreesize,
		      ancestry,
		      nmarkers,
		      nalleles,
		      haplabel,
		      spacesize,
		      maf,
		      ppotential);
    if(c.debug.par) DebugCreateMarkerPPots(npedigrees,
				           pedigreesize,
				           pedigree,
				           nmarkers,
				           nalleles,
				           haplabel,
				           spacesize,
				           ppotential);

    if(lod)
    {
       CreateDiseasePPots(npedigrees,
			  pedigreesize,
			  ancestry,
			  haplabel,
			  nloci,
			  disaf,
			  ppotential);
  
       if(c.debug.par) DebugCreateDiseasePPots(npedigrees,
                        	               pedigreesize,
                        	               pedigree,
                        	               haplabel,
                        	               nloci,  
                        	               ppotential);
    }
    Peel(npedigrees,
         pedigreesize,
         ancestry,
         peelorder,
         peeltype,
         peelinto,
         nloci,
         c.processorder.par[0]-1,
         spacesize,
         jpotflag,
         haplabel,
         ibdobs,
         ibdF,
         ibdM,
         ppotential,
         jpotential,
         wppotential,
         &prob[FIRSTMODEL][0]);

    if(c.debug.par) DebugPeel(npedigrees,
                              pedigreesize,
                              pedigree,
                              c.processorder.par[0]-1,
                              peelinto,
                              jpotflag,
                              spacesize,
                              haplabel,
                              wppotential,
                              jpotential);

/* I did this because the weight ibdobsprob was 0 forcing the
weight to be 0 for a pedigree that I was lookin at 
    if(nloci==1)
      weight[0] = CalcWeight(npedigrees, pedigreesize, peelorder, spacesize, c.processorder.par[0]-1, wppotential);
    else */
      weight[0] = CalcWeight(npedigrees, pedigreesize, peelorder, spacesize, c.processorder.par[0]-1, wppotential);
    weight[0] *= ibdobsprob;

  }
  else {/* complex pedigree */
    GetWPP(npedigrees,
           ancestry,
           haplabel,
           nloci,
           lod, 
           affected,
           liability,
	   disaf,
	   maf,
           npeels,
           peelset,
           peelspace,
           peelspaces,
           whaps,
           wpp);
    if(c.debug.par) DebugGetWPP(npedigrees,
				nloci,
				npeels,
                                peelspace,
                                wpp);


    /****************************
    *  Get indices for peeling  * 
    ****************************/
    GetTransIndex(npedigrees,
                  npeels,
                  cutset,
                  peelset,
                  ancestry,
                  &ntrans,
                  &tidx);
    if(c.debug.par) DebugGetTransIndex(npedigrees,
                                       npeels,
                                       ntrans,
                                       tidx);

    GetRIndex(npedigrees,
              npeels,
              cutset,
              peelset,
              &nrinto,
              &ridx); 
    if(c.debug.par) DebugGetRIndex(npedigrees,
                                   npeels,
                                   nrinto,
                                   ridx);

/*
    ntrans = (int **) calloc(npedigrees, sizeof(int *));
    ntrans[0] = (int *) calloc(pedigreesize[0], sizeof(int));
    ntrans[0][0]=2;
    ntrans[0][1]=0;
    ntrans[0][2]=0;

    tidx = (tindex ***) calloc(npedigrees, sizeof(tindex **));
    tidx[0] = (tindex **) calloc(npeels[0], sizeof(tindex*));
    tidx[0][0] = (tindex *) calloc(ntrans[0][0], sizeof(tindex));

    tidx[0][0][0].set[0]=2;
    tidx[0][0][0].set[2]=0;
    tidx[0][0][0].set[1]=1;

    tidx[0][0][0].ped[0]=2;
    tidx[0][0][0].ped[2]=0;
    tidx[0][0][0].ped[1]=1;

    tidx[0][0][1].set[0]=3;
    tidx[0][0][1].set[2]=0;
    tidx[0][0][1].set[1]=1;

    tidx[0][0][1].ped[0]=3;
    tidx[0][0][1].ped[2]=0;
    tidx[0][0][1].ped[1]=1;


    nrinto = (int **) calloc(npedigrees, sizeof(int *));
    nrinto[0] = (int *) calloc(pedigreesize[0], sizeof(int));
    nrinto[0][0]=0;
    nrinto[0][1]=1;
    nrinto[0][2]=1;

    ridx = (Rindex ***) calloc(npedigrees, sizeof(Rindex **));
    ridx[0] = (Rindex **) calloc(npeels[0], sizeof(Rindex *));
    ridx[0][1] = (Rindex *) calloc(nrinto[0][1], sizeof(Rindex));
    ridx[0][2] = (Rindex *) calloc(nrinto[0][1], sizeof(Rindex));
    ridx[0][1][0].n=2;
    ridx[0][2][0].n=1;
    ridx[0][1][0].set = (int *) calloc(ridx[0][1][0].n, sizeof(int));
    ridx[0][2][0].set = (int *) calloc(ridx[0][2][0].n, sizeof(int));

    ridx[0][1][0].r=0;
    ridx[0][1][0].set[0]=0;
    ridx[0][1][0].set[1]=1;

    ridx[0][2][0].r=1;
    ridx[0][2][0].set[0]=0;
*/
    ComplexPeel(npedigrees,
		ntrans,
		nrinto,
                tidx,
                ridx,
                haplabel,
		ibdobs,
		ibdF,
		ibdM,
		&prob[FIRSTMODEL][0],
                nloci,
                c.processorder.par[0]-1,
                npeels,
                cutset,
                &cutspace[c.processorder.par[0]-1][0],
                &cutspaces[c.processorder.par[0]-1][0],
                peelset,
                &peelspace[c.processorder.par[0]-1][0],
                &peelspaces[c.processorder.par[0]-1][0],
                whaps,
                whaps2,
                &wpp[c.processorder.par[0]-1][0],
		&rfunction[c.processorder.par[0]-1][0]);
    if(c.debug.par) DebugComplexPeel(npedigrees,
                       		     npeels,
				     &cutspace[c.processorder.par[0]-1][0],
                                     &rfunction[c.processorder.par[0]-1][0]);

    weight[0] = ComplexCalcWeight(npedigrees, npeels, &rfunction[c.processorder.par[0]-1][0]);

  }


  for(iter=0; iter<c.maxiter.par; iter++) {

    CopyIBD(npedigrees,
  	    pedigreesize,
	    nloci,
	    ibdobs,
	    ibdsim);

    if(!c.peel.par)  /* not complex */
      Simulate(npedigrees,
               pedigreesize,
               ancestry,
               spacesize,
               jpotflag,
               peelorder,
               peeltype,
               peelinto,
               haplabel,
               nloci,
               c.processorder.par[0]-1,
               ibdsim,
               ibdF,
               ibdM,
               wppotential,
               simwppotential,
               jpotential,
               simjpotential,
               simhap,
               shap,
               &prob[FIRSTMODEL][0]);

    else
      ComplexSimulate(npedigrees,
                      ntrans,
                      nrinto,
                      tidx,
                      ridx,
                      haplabel,
                      ibdsim,
                      ibdF,
                      ibdM,
                      &prob[FIRSTMODEL][0],
                      nloci,
                      c.processorder.par[0]-1,
                      npeels,
                      cutset,
                      &cutspace[c.processorder.par[0]-1][0],
                      &cutspaces[c.processorder.par[0]-1][0],
                      peelset,
                      &peelspace[c.processorder.par[0]-1][0],
                      &peelspaces[c.processorder.par[0]-1][0],
                      whaps,
                      whaps2,
                      &wpp[c.processorder.par[0]-1][0],
                      &shap[c.processorder.par[0]-1][0],
                      simhap,
                      simprob,
                      &rfunction[c.processorder.par[0]-1][0]);

    if (c.debug.par) DebugSimulate(npedigrees,
                                   pedigreesize,
                                   pedigree,
                                   simhap,
                                   c.processorder.par[0]-1);


    UpdateIBDsim(npedigrees,
                 pedigreesize,
                 ancestry,
                 c.processorder.par[0]-1,
                 simhap,
                 ibdsim);

    if(c.debug.par) DebugGetIBDObs(npedigrees,
                                   pedigreesize,
                                   pedigree,
                                   nloci,
                                   ibdsim);
   if(!c.peel.par) /* modify for complex remove  */
      SimBlank(npedigrees,
               pedigreesize,
               c.processorder.par[0]-1,
               blank,
               ibdsim); 

    /* for(locus=1; locus < nloci-1; locus++) { */
    for(locus=1; locus < nmarkers; locus++) {

      if(!c.peel.par){  /* not complex */
        Peel(npedigrees,
	     pedigreesize,
	     ancestry,
	     peelorder,
	     peeltype,
	     peelinto,
	     nloci,
	     c.processorder.par[locus]-1,
	     spacesize,
	     jpotflag,
	     haplabel,
	     ibdsim,
	     ibdF,
	     ibdM,
	     ppotential,
	     jpotential,
	     wppotential,
	     &prob[FIRSTMODEL][0]);
        if(c.debug.par) DebugPeel(npedigrees,
               		          pedigreesize,
               		          pedigree,
               		          c.processorder.par[locus]-1,
               		          peelinto,
               		          jpotflag,
               		          spacesize,
               		          haplabel,
               		          wppotential,
               		          jpotential);

        weight[locus] = CalcWeight(npedigrees, pedigreesize, peelorder, spacesize, c.processorder.par[locus]-1, wppotential);


        Simulate(npedigrees,
                 pedigreesize,
	         ancestry,
                 spacesize,
                 jpotflag,
                 peelorder,
                 peeltype,
                 peelinto,
                 haplabel,
	         nloci,
                 c.processorder.par[locus]-1,
	         ibdsim,
	         ibdF,
	         ibdM,
                 wppotential,
                 simwppotential,
                 jpotential,
                 simjpotential,
                 simhap,
                 shap,
                 &prob[FIRSTMODEL][0]);

      }
      else {
        ComplexPeel(npedigrees,
		    ntrans,
		    nrinto,
                    tidx,
                    ridx,
                    haplabel,
		    ibdsim,
		    ibdF,
		    ibdM,
		    &prob[FIRSTMODEL][0],
                    nloci,
                    c.processorder.par[locus]-1,
                    npeels,
                    cutset,
                    &cutspace[c.processorder.par[locus]-1][0],
                    &cutspaces[c.processorder.par[locus]-1][0],
                    peelset,
                    &peelspace[c.processorder.par[locus]-1][0],
                    &peelspaces[c.processorder.par[locus]-1][0],
                    whaps,
                    whaps2,
                    &wpp[c.processorder.par[locus]-1][0],
		    &rfunction[c.processorder.par[locus]-1][0]);
        if(c.debug.par) DebugComplexPeel(npedigrees,
                       		         npeels,
				         &cutspace[c.processorder.par[locus]-1][0],
                                         &rfunction[c.processorder.par[locus]-1][0]);

        ComplexSimulate(npedigrees,
                        ntrans,
                        nrinto,
                        tidx,
                        ridx,
                        haplabel,
                        ibdsim,
                        ibdF,
                        ibdM,
                        &prob[FIRSTMODEL][0],
                        nloci,
                        c.processorder.par[locus]-1,
                        npeels,
                        cutset,
                        &cutspace[c.processorder.par[locus]-1][0],
                        &cutspaces[c.processorder.par[locus]-1][0],
                        peelset,
                        &peelspace[c.processorder.par[locus]-1][0],
                        &peelspaces[c.processorder.par[locus]-1][0],
                        whaps,
                        whaps2,
                        &wpp[c.processorder.par[locus]-1][0],
                        &shap[c.processorder.par[locus]-1][0],
                        simhap,
                        simprob,
                        &rfunction[c.processorder.par[locus]-1][0]);

        weight[locus] = ComplexCalcWeight(npedigrees, npeels, &rfunction[c.processorder.par[locus]-1][0]);
      }

      if (c.debug.par) DebugSimulate(npedigrees,
                 	             pedigreesize,
                   	             pedigree,
                   	             simhap,
                   	             c.processorder.par[locus]-1);
  
      UpdateIBDsim(npedigrees,
	           pedigreesize,
	           ancestry,
	           c.processorder.par[locus]-1,
	           simhap,
	           ibdsim);
  
      if(c.debug.par) DebugGetIBDObs(npedigrees,
                                     pedigreesize,
                                     pedigree,
                                     nloci,
                                     ibdsim);
    
      if(!c.peel.par) 
        SimBlank(npedigrees,
	         pedigreesize,
                 c.processorder.par[locus]-1,
                 blank,
                 ibdsim); 

    }

    if(lod) /* 6/11/02 */
    for(model=0; model < c.file.nmodels; model++) 
    {
      IBDReset(npedigrees,
	       pedigreesize,
	       c.processorder.par[nloci-1]-1,
               ibdsim,
	       ibdobs);

      if(!c.peel.par){  /* not complex */
        Peel(npedigrees,
             pedigreesize,
             ancestry,
             peelorder,
             peeltype,
             peelinto,
             nloci,
             c.processorder.par[nloci-1]-1,
             spacesize,
             jpotflag,
             haplabel,
             ibdsim,
             ibdF,
             ibdM,
             ppotential,
             jpotential,
             wppotential,
             &prob[model][0]);
        if(c.debug.par) DebugPeel(npedigrees,
                                  pedigreesize,
                                  pedigree,
                                  c.processorder.par[nloci-1]-1,
                                  peelinto,
                                  jpotflag,
                                  spacesize,
                                  haplabel,
                                  wppotential,
                                  jpotential);

        weight[nloci+model-1] = CalcWeight(npedigrees, pedigreesize, peelorder,spacesize, c.processorder.par[nloci-1]-1, wppotential);
                              

        Simulate(npedigrees,
                 pedigreesize,
                 ancestry,
                 spacesize,
                 jpotflag,
                 peelorder,
                 peeltype,
                 peelinto,
                 haplabel,
                 nloci,
                 c.processorder.par[nloci-1]-1,
                 ibdsim,
                 ibdF,
                 ibdM,
                 wppotential,
                 simwppotential,
                 jpotential,
                 simjpotential,
                 simhap,
                 shap,
                 &prob[model][0]);
      }
      else {
        ComplexPeel(npedigrees,
		    ntrans,
		    nrinto,
                    tidx,
                    ridx,
                    haplabel,
		    ibdsim,
		    ibdF,
		    ibdM,
		    &prob[model][0],
                    nloci,
                    c.processorder.par[locus]-1,
                    npeels,
                    cutset,
                    &cutspace[c.processorder.par[nloci-1]-1][0],
                    &cutspaces[c.processorder.par[nloci-1]-1][0],
                    peelset,
                    &peelspace[c.processorder.par[nloci-1]-1][0],
                    &peelspaces[c.processorder.par[nloci-1]-1][0],
                    whaps,
                    whaps2,
                    &wpp[c.processorder.par[nloci-1]-1][0],
		    &rfunction[c.processorder.par[nloci-1]-1][0]);
        if(c.debug.par) DebugComplexPeel(npedigrees,
                       		         npeels,
				         &cutspace[c.processorder.par[nloci-1]-1][0],
                                         &rfunction[c.processorder.par[nloci-1]-1][0]);

        weight[nloci+model-1] = ComplexCalcWeight(npedigrees, npeels, &rfunction[c.processorder.par[nloci-1]-1][0]);

        ComplexSimulate(npedigrees,
                        ntrans,
                        nrinto,
                        tidx,
                        ridx,
                        haplabel,
                        ibdsim,
                        ibdF,
                        ibdM,
                        &prob[model][0],
                        nloci,
                        c.processorder.par[nloci-1]-1,
                        npeels,
                        cutset,
                        &cutspace[c.processorder.par[nloci-1]-1][0],
                        &cutspaces[c.processorder.par[nloci-1]-1][0],
                        peelset,
                        &peelspace[c.processorder.par[nloci-1]-1][0],
                        &peelspaces[c.processorder.par[nloci-1]-1][0],
                        whaps,
                        whaps2,
                        &wpp[c.processorder.par[nloci-1]-1][0],
                        &shap[c.processorder.par[nloci-1]-1][0],
                        simhap,
                        simprob,
                        &rfunction[c.processorder.par[nloci-1]-1][0]);
      }
      if (c.debug.par) DebugSimulate(npedigrees,
                                     pedigreesize,
                                     pedigree,
                                     simhap,
                                     c.processorder.par[nloci-1]-1);

      UpdateIBDsim(npedigrees,
                   pedigreesize,
                   ancestry,
                   c.processorder.par[nloci-1]-1,
                   simhap,
                   ibdsim);

      if(c.debug.par) DebugGetIBDObs(npedigrees,
                                     pedigreesize,
                                     pedigree,
                                     nloci,
                                     ibdsim);

      if(!c.peel.par) 
        SimBlank(npedigrees,
                 pedigreesize,
                 c.processorder.par[nloci-1]-1,
                 blank,
                 ibdsim); 

      for(k=0; k<numlik[model]; k++)
	likratio[model][k] = CalcLikRatio(npedigrees,
				          pedigreesize,
				          ibdsim,
				          ibdF,
				          ibdM,
				          nloci,
				          model,
				          k,
				          prob,
				          likprob);

    }


    printf("%5d ",iter+1);
/* this is a hack */
/*    if(nmarkers==1 && !lod && c.file.def) 
      printf("%le ",weight[0]); */
    for(i=0; i<nloci+c.file.nmodels-1-!lod; i++) 
      printf("%le ",weight[i]);


    if((strcmp(c.analysis.par,"NPL")==0)
      || (strcmp(c.analysis.par,"BOTH")==0)
      || c.hasemanelston.par==TRUE) 
		      DumpIBDS(npedigrees,
                               pedigreesize,
	                       ancestry,
                               nmarkers,
                               ibdsim,
	                       &locusorder[0][0]); 

    if(c.dumphap.par)
      DumpHaps(npedigrees,
               pedigreesize,
               simhap,
               nmarkers,
               nalleles,
               hap);

/* this is a hack */
    /* if(nmarkers==1 && !lod && c.file.def) 
        printf("%le ",1.0); */
    if(lod)
      for(i=0; i<c.file.nmodels; i++)
        for(k=0; k<numlik[i]; k++)
          printf("%le ",likratio[i][k]);
    else 
      printf("%le ",1.0); 
    printf("\n");

    fflush(stdout);
  }

  DumpCommands(c,
	       argv,
	       argc,
	       npedigrees,
	       nloci);

  PlantSeed();
  exit(0);

}
