/* ********************************************************************

    Program:   NPL
    File:      CreateDistances.LR.c
    Author:    Zachary Skrivanek
    Date:      May 23, 2000
    Purpose:   Creates array of distances for NPL.  This is slightly 
               different than the version for SIMPLE in that is also
               calculates the statistic directly at the markers without
               using re-weighting.  We assume that there is a "disease".
    Modified:  July 2, 2002.  Estimated over marker, not just interior
               points.

********************************************************************* */

#include <stdlib.h> /* exit */

void  CreateDistances_LR(int     nmarkers, 
		         int     *nmodels,
                         int     nsteps,
		         double  *mdistances,
		         int     ***locusorder,
		         double  ***distances,
		         int     **numlik,
		         double  ****likdistances)

{

  int    i,j,k,l;
  int    nloci;
  int    **tlocusorder;
  int    *tnumlik;
  int    tnmodels, model;
  double **tdistances;
  double ***tlikdistances;

  nloci = nmarkers + 1;
  nsteps -= 1;  /* to conform to GH */
  tnmodels = 2*nmarkers;

  /*******************
  ** allocate space **
  *******************/
  tlocusorder   = (int **) calloc(tnmodels,sizeof(int *));
  tnumlik       = (int *) calloc(tnmodels,sizeof(int));       
  tdistances    = (double **) calloc(tnmodels,sizeof(double *));
  tlikdistances = (double ***) calloc(tnmodels,sizeof(double **));


  for(i=0; i < tnmodels; i++)
  {
    /* we give the null case xtra space in case we want to
    convert to rec fractions and use mark's algorithm to 
    calculate ibd probs in which case we tack on a rec. frac
    of .5 for the null disease case (independent) */
    tlocusorder[i]   = (int *) calloc(nloci,sizeof(int));
    tdistances[i]    = (double *) calloc(nloci-1,sizeof(double));
    if(i%2 == 0) tnumlik[i] = nsteps+2;/* the null case plus markers */
    else tnumlik[i] = 1;
    tlikdistances[i] = (double **) calloc(tnumlik[i],sizeof(double *));
    for(j=0; j < tnumlik[i]; j++)
      tlikdistances[i][j] = (double *) calloc(nloci-1,sizeof(double));
  }
  tnumlik[0]=1;

  /***********************
  ** create tlocusorder ** 
  ***********************/

  for(i=0; i < nloci; i++){
    tlocusorder[0][i] = (i+1);
    tlocusorder[tnmodels-1][i] = (i+1);
  }

  model=1;
  for(i=1; i <= nmarkers-1; i++){
    for(j=1; j<= nmarkers; j++){
      if(j <= i){
	tlocusorder[model][j-1]=j;
	tlocusorder[model+1][j-1]=j;
      }
      else if(j==i+1){
	tlocusorder[model][j-1]=nloci;
	tlocusorder[model+1][j-1]=nloci;
	tlocusorder[model][j]=j;
	tlocusorder[model+1][j]=j;
      }	
      else{
	tlocusorder[model][j]=j;
	tlocusorder[model+1][j]=j;
      }	
    }
    model += 2;
  }

  /********************************
  ** take care of the null model **
  *********************************/
  
  for(i=0; i<(nmarkers-1); i++){
    tdistances[0][i] = mdistances[i];
    tlikdistances[0][0][i] = mdistances[i];
  }

  /**************************************
  ** create distances for disease case **
  **************************************/

  /* do models on top of markers */
  for(i=0; i < nmarkers; i++){
    model = 2*i+1;
    tdistances[model][i]=0.0;
    tlikdistances[model][0][i]=0.0;
    for(j=0; j<nmarkers-1; j++){
      if(j<i) {
	tdistances[model][j]=mdistances[j];
	tlikdistances[model][0][j]=mdistances[j];
      }
      else {
	tdistances[model][j+1]=mdistances[j];
	tlikdistances[model][0][j+1]=mdistances[j];
      }
    }
  }

  /* do the rest of the distances */
  for(i=0; i<nmarkers-1; i++){
    model=2*(i+1);
    for(j=0; j<nmarkers-1; j++){
      if(j < i)
	tdistances[model][j]=mdistances[j];
      else if(j==i){
	tdistances[model][j]=mdistances[j]/2.0;
	tdistances[model][j+1]=mdistances[j]/2.0;
      }
      else       
	tdistances[model][j+1]=mdistances[j];
    }
  }	

  /* do the rest of the likdistances */
  for(i=0; i < nmarkers - 1; i++){
    model=2*(i+1); 
    for(j=0; j < tnumlik[model]; j++)/* 7/2/02 Zak */
      for(k=0; k< (nmarkers - 1); k++){
        if(k < i ) 
          tlikdistances[model][j][k]=mdistances[k]; /* 7/2/02 Zak */
        else if(k == i ){
          tlikdistances[model][j][k]=(mdistances[k] *  ( (double) j) / ( (double) (tnumlik[model]) - 1)); /* 7/2/02 Zak */
     
          tlikdistances[model][j][k+1]=(mdistances[k] - tlikdistances[model][j][k]);
        }
        else 
          tlikdistances[model][j][k+1]=mdistances[k];
      }
  } 



  *locusorder   = tlocusorder;
  *distances    = tdistances;
  *numlik       = tnumlik;
  *likdistances = tlikdistances;
  *nmodels      =tnmodels;
}


void DebugCreateDistances_LR(int     nmarkers,
		             int     nmodels,
		             int     **locusorder,
		             double  **distances,
		             int     *numlik,
		             double  ***likdistances)
{
  char response[2],c;

  int nloci;

  nloci=nmarkers+1;

  printf("\n\nDebug CreateDistances()? => ");
  scanf("%s", response);
  if(strcmp(response, "q") == 0) exit(0);
  if(strcmp(response, "n") == 0) return;
  if(strcmp(response, "y") == 0) 
  {

    int    i,j,k;
    int    start,count;
    div_t  nloops;

    /* calculate number of rows needed to print out distances */
    nloops = div(nloci, 7);

    for(i=0; i<nmodels; i++) 
    {
      printf("\n\n");
      printf("\n    s\n    u\n    b\nm   m\no   o\nd   d\ne   e\n");
      printf("l   l    Locus Order/ Genetic Distances\n");
      printf("*** *** ******************************************"); 
      printf("***********************\n\n");
      /* print locus order */
      /* start loop is used so we don't print off screen */
      for(start=0; start <= nloops.quot; start++) 
      {
	count=0;
	if( (nloci - (start * 7)) != 1)
	{
          for(j=(start * 7); j<nloci ;j++)
	  {
	    if(j==(start * 7)) printf("order:  ");
/*	    if(i==0 && j==(nloci-1))
	      printf("dis");
	    else if(locusorder[i][j]==nloci ) */
	    if(locusorder[i][j]==nloci ) 
	      printf("dis      ");
            else printf("%-3d      ",locusorder[i][j]);
            count++;
	    if( count%8 == 0 ) break; 
          }
	  printf("\n");

	  /* print underline */
	  count=0;
          for(j=(start * 7); j<(nloci-1);j++ )
	  {
	    if(j==(start * 7)) printf("        ");
            printf("=========");
	    count++;
	    if( count%8 == 0 ) break;
          }
	  printf("\n");
        } /* if statement */
 
        /* print distances */ 
        for(j=(start * 7); j<(nloci-1);)
	{
          if(j==(start * 7)) printf("%-3d     ",i); 
          if(i==0 && j==(nloci-2)) 
	    printf("   infinity");
          else printf("   %6.4lf",distances[i][j]);
	  j++;
	  if( j%7 == 0 )
	  {
	    printf("\n");
	    break;
          }
        }
        if(j==(nloci-1) && j%7 > 0)
	  printf("\n");
      }
      printf("--- ---    ------\n");

      for(k=0; k < numlik[i]; k++) 
      {
        for(start=0; start <= nloops.quot; start++)
        {
          for(j=(start * 7); j<(nloci-1);)
	  {
            if(j==(start * 7)) printf("%-3d %-3d ",i,k); 
            if(i==0 && j==(nloci-2) ) 
	      printf("   infinity");
            else printf("   %-6.4lf",likdistances[i][k][j]);
	    j++;
	    if( j%7 == 0 )
	    {
	      printf("\n");
	      break;
            }
          }
          if(j==(nloci-1) && j%7 > 0)
	    printf("\n");
        }
      }
      printf("\npress 'return' to continue");
      scanf("%c",&c);
      printf("\n");
    }
  }
}
