/***************************************************************

  Program: scan 
  File:    EM.c
  Author:  Zachary Skrivanek 
  Date:    February 3, 2001 
  Purpose: Use EM algorithm to get ML estimates of regression
	   parameters for haseman elston statistic.

***************************************************************/
#include "simple.h"
#include <math.h> /* pow, sqrt */
#include <stdlib.h> /* pow, sqrt */
#include <stdio.h> /* fprintf */

#define TRUE 1
#define FALSE 0
#define stop 200 /* when to kill the EM loop for nonconvergence */ 
#define delta 0.0000001 /* when to kill the EM loop for convergence */ 
#define ALPHA 0 
#define BETA 1 
#define SIGMA 2 
#define PI 3.1415926535897932384626433832795028841972




void  EM(int    npedigrees,
	 prs    ***pairs,
	 int    **npairs,
	 int    nmodels,
	 int    *numlik,
	 double ******IBDprob,
	 double ****line){

  int    converge;
  double a1, b1, s1_2, a2, b2, s2_2;
  double v[3];
  double lik;
  int    i, j, pair, ped, type, ibd, niter; /* counter */
  double sum_v, sum_y=0.0, sum_vy, sum_v_2, sum_y_2, y;
  double den, prob_v_x, prob_y_vx, tmp;
  int    n;

  char   *labels[] ={"sib"};

  double abs_value(double);
  double norm_pdf(double, double, double);

  v[0]=0.0;
  v[1]=0.5;
  v[2]=1.0;


  for(i=1; i<nmodels; i++)
    for(j=0; j<numlik[i]; j++)
      /* for(type=0; type<ntypes; type++){ 7/18/02 */
      type=0;
        n=0;
        sum_y=0.0;
        sum_y_2=0.0;
        for(ped=0; ped<npedigrees; ped++){
          for(pair=0; pair<npairs[type][ped]; pair++){
            y = pairs[type][ped][pair].pheno;
            sum_y += y; 
            sum_y_2 += pow(y,2.0);
            n++;
          } /* pair */
        } /* ped */
      /* default values of parameters */
        a1 = sum_y/ ((double) n);
        b1 = 0.0;
        s1_2 = sum_y_2-pow(sum_y, 2.0)/ ((double) n);
        s1_2 /= ((double) n);

        niter=0;
        converge=FALSE;
        while(converge==FALSE){
          if(niter>stop) {
	    fprintf(stderr,"type %s model %d, numlik %d\n", labels[type], i, j);
	    nerror("EM algorithm did not converge");
          }

          sum_v=0.0;
          sum_v_2=0.0;
          sum_vy=0.0;
        /* E step */
          for(ped=0; ped<npedigrees; ped++){
	    for(pair=0; pair<npairs[type][ped]; pair++){
	      den=0.0;
	      y=pairs[type][ped][pair].pheno;
	      for(ibd=0; ibd<3; ibd++){
	        prob_v_x = IBDprob[i-1][j][type][ped][pair][ibd];
	        prob_y_vx = norm_pdf(y, a1+b1*v[ibd], s1_2);
	        den+=prob_v_x*prob_y_vx;
              }
	      for(ibd=0; ibd<3; ibd++){
	        prob_v_x=IBDprob[i-1][j][type][ped][pair][ibd];
		prob_y_vx=norm_pdf(y,a1+b1*v[ibd],s1_2);
		sum_v += v[ibd]*prob_v_x*prob_y_vx/den;
		sum_v_2 += pow(v[ibd],2.0)*prob_v_x*prob_y_vx/den;
		sum_vy += v[ibd]*y*prob_v_x*prob_y_vx/den;
              } /* ibd */
            } /* pair */
          } /* ped */ /* E step */

        /* M step */
	  b2 = sum_vy-sum_v*sum_y/((double) n);
	  if((sum_v_2 - pow(sum_v, 2.0)/((double) n))==0.0)
	    nerror("EM:  singular matrix!");
	  b2 /= (sum_v_2 - pow(sum_v, 2.0)/((double) n));
	  a2 = sum_y / ((double) n) - b2*sum_v/((double) n);
	  s2_2 = sum_y_2 - 2*a2*sum_y-2*b2*sum_vy+((double) n)*pow(a2,2.0)+ 
		 2*a2*b2*sum_v + pow(b2,2.0)*sum_v_2;
	  s2_2 /= ((double) n);

          if(niter%10==0){ 
	    fprintf(stderr,"%d %d %d a1=%lf a2=%lf\n", i,j,niter,a1,a2);
	    fprintf(stderr,"%d %d %d b1=%lf b2=%lf\n", i,j,niter,b1,b2);
	    fprintf(stderr,"%d %d %d s1_2=%lf s2_2=%lf\n", i,j,niter,s1_2,s2_2);
	    /* this is actually the Q function */
	    /* lik=-((double) n)/2*log(2*PI*s2_2)-sum_y_2/(2*s2_2) + a2*sum_y/s2_2 + b2*sum_vy/s2_2-((double) n)*pow(a2,2.0)/(2*s2_2)-a2*b2*sum_v/s2_2-pow(b2,2.0)*sum_v_2/(2*s2_2);	
	    fprintf(stderr,"Q = %lf\n",lik); */
	    lik=1.0;
            for(ped=0; ped<npedigrees; ped++)
	      for(pair=0; pair<npairs[type][ped]; pair++){
		tmp=0.0;
	        for(ibd=0; ibd<3; ibd++){
	          prob_v_x = IBDprob[i-1][j][type][ped][pair][ibd];
	          prob_y_vx = norm_pdf(y, a2+b2*v[ibd], s2_2);
	          tmp += prob_v_x*prob_y_vx;
                }
		lik *= tmp;
              }
	    fprintf(stderr,"loge(lik) = %lf\n",log(lik)); 
          } 

        /* check for convergence */
          niter++;
	  if(abs_value(a1-a2) < delta && 
	     abs_value(b1-b2) < delta && 
	     abs_value(s1_2-s1_2) < delta ) {
	    converge=TRUE; 
	    fprintf(stderr,"%d %d %d a1=%lf a2=%lf\n", i,j,niter,a1,a2);
	    fprintf(stderr,"%d %d %d b1=%lf b2=%lf\n", i,j,niter,b1,b2);
	    fprintf(stderr,"%d %d %d s1_2=%lf s2_2=%lf\n", i,j,niter,s1_2,s2_2);
	   /* lik=-((double) n)/2*log(2*PI*s2_2)-sum_y_2/(2*s2_2) + a2*sum_y/s2_2 + b2*sum_vy/s2_2-((double) n)*pow(a2,2.0)/(2*s2_2)-a2*b2*sum_v/s2_2-pow(b2,2.0)*sum_v_2/(2*s2_2);	
	    fprintf(stderr,"Q = %lf\n",lik);*/
	    lik=1.0;
            for(ped=0; ped<npedigrees; ped++)
	      for(pair=0; pair<npairs[type][ped]; pair++){
		tmp=0.0;
	        for(ibd=0; ibd<3; ibd++){
	          prob_v_x = IBDprob[i-1][j][type][ped][pair][ibd];
	          prob_y_vx = norm_pdf(y, a2+b2*v[ibd], s2_2);
	          tmp += prob_v_x*prob_y_vx;
                }
		lik *= tmp;
              }
	    fprintf(stderr,"loge(lik) = %lf\n",log(lik)); 
          }
          else{
	    a1=a2;
	    b1=b2;
	    s1_2=s2_2;
          }
        } /* converge loop */ 
	line[i-1][j][type][ALPHA]=a2; 
	line[i-1][j][type][BETA]=b2; 
	line[i-1][j][type][SIGMA]=s2_2;
     /*  }7/18/02 type */
}
  
	         
