/***************************************************************

  Program: SIMPLE 
  File:    ReadPeelFiles.c
  Author:  Zachary Skrivanek 
  Date:    January 26, 2002 
  Purpose: Read in peel files for complex peeling.  We assume
	   that the peel files are formatted with two columns
	   corresponding to the cut sets and peel sets in peel-
	   ing order.  The members of each set are separated
	   by a ',' with the ',' immediately following each pid.
	   The pids are assumed to be integers.  The last cut
	   set is always a singulare '0' since this is the root.
	   For example:

	   2,3	1
	   2	3
	   0	2

***************************************************************/

#include <stdio.h> /*fclose */
#include "simple.h" /* c */

#define MAX 256

count   i,j,ped,peel;
sindex  *c,*p;

void  ReadPeelFiles(char    **PeelName,
		    int     npedigrees,
		    int     *pedigreesize,
		    int     **npeels,
		    sindex  ***cutset, 
		    sindex  ***peelset){

  FILE   *peelfile, *gfopen();
  char   buffer[MAX+1],tc;
  int    tmp;

  sindex  **tcutset, **tpeelset;
  int     *tnpeels;

  void    error();

  tnpeels = (int *) calloc(npedigrees, sizeof(int));
  tcutset = (sindex **) calloc(npedigrees, sizeof(sindex *));
  tpeelset = (sindex **) calloc(npedigrees, sizeof(sindex *));

  for(ped=0; ped < npedigrees; ped++){
    peelfile = gfopen(PeelName[ped],"r");
  /* count number of peels.  This is determined */
  /* by number of lines that begin with and int */
    tnpeels[ped]=0;
    while(fgets(buffer, MAX, peelfile)!=NULL) {
      if(sscanf(buffer,"%d",&tmp)==0) break;
      tnpeels[ped]++;
    }

    tcutset[ped] = (sindex *) calloc(tnpeels[ped], sizeof(sindex));
    tpeelset[ped] = (sindex *) calloc(tnpeels[ped], sizeof(sindex));
  /* find out how many are in each cut set and peel set */
    rewind(peelfile);
    for(peel=0; peel<tnpeels[ped]; peel++){
     
    /* get cut set */
      tcutset[ped][peel].n=0;
      while(fscanf(peelfile,"%d%[,]",&tmp,&tc)==2) tcutset[ped][peel].n++;
      tcutset[ped][peel].n++; 

    /* get peel set */
      tpeelset[ped][peel].n=0;
      while(fscanf(peelfile,"%d%[,]",&tmp,&tc)==2) tpeelset[ped][peel].n++;
      tpeelset[ped][peel].n++;
    }

    rewind(peelfile);
  /*  Read in cut and peel sets */
    for(peel=0; peel<tnpeels[ped]; peel++){

    /* for reading ease */
      c = &tcutset[ped][peel];
      p = &tpeelset[ped][peel];

      tcutset[ped][peel].pid = (int *) calloc(c->n, sizeof(int));
      tcutset[ped][peel].ped = (int *) calloc(c->n, sizeof(int));
      tpeelset[ped][peel].pid = (int *) calloc(p->n, sizeof(int));
      tpeelset[ped][peel].ped = (int *) calloc(p->n, sizeof(int));
     
      for(i=0; i < c->n-1; i++)
      /* the first n-1 will be immediately followed by a ',' */
	if(fscanf(peelfile,"%d,",&c->pid[i])!=1)
	  error("Cut",i+1,peel+1,PeelName[ped]);
    /* the last member has no ',' after it */
      if(fscanf(peelfile,"%d",&c->pid[c->n-1])!=1)
	error("Cut",i+1,peel+1,PeelName[ped]);

      for(i=0; i < p->n-1; i++)
      /* the first n-1 will be immediately followed by a ',' */
	if(fscanf(peelfile,"%d,",&p->pid[i])!=1)
	  error("Peel",i+1,peel+1,PeelName[ped]);
    /* the last member has no ',' after it */
      if(fscanf(peelfile,"%d",&p->pid[p->n-1])!=1)
	error("Peel",i+1,peel+1,PeelName[ped]);
    }

  /* check that sum of peel sets = pedigreesize */
    tmp=0;
    for(peel=0; peel<tnpeels[ped]; peel++)
      tmp += tpeelset[ped][peel].n;
    
    if(tmp != pedigreesize[ped]){
      fprintf(stderr,"%d members in pedigree %d, but %d members in peel file %s\n",pedigreesize[ped],ped+1,tmp,PeelName[ped]);
      nerror("Format error in peel file!");
    }
    /* set the last cut set size to 0 */
    /* since we are peeling the root */
    /* we allocated space and input a */
    /* pids for this set but they should */
    /* be 0 and ow meaningless */
    tcutset[ped][tnpeels[ped]-1].n=0;
  }

  *cutset = tcutset;
  *peelset = tpeelset;
  *npeels = tnpeels;

}
     
void error(char  *set,
	   int   mem,
	   int   peel,
	   char  *file){

  fprintf(stderr,"%s member %d in peel %d of peel file %s should be an integer\n",set,mem,peel,file);
  nerror("Format error in peel file!");
}

void  DebugReadPeelFiles(int     npedigrees,
		         int     *npeels,
		         sindex  **cutset, 
		         sindex  **peelset){

  char response[2];

  printf("\n\nDebug ReadPeelFiles()? => ");
  scanf("%s", response);
  if(strcmp(response, "q") == 0) exit(0);
  if(strcmp(response, "y") == 0)
  {
    printf("\n\n");
    for(ped=0; ped < npedigrees; ped++){
      printf("ped %2d\n",ped+1);
      printf("******\n");
      for(peel=0; peel<npeels[ped]; peel++){
	c = &cutset[ped][peel];
	p = &peelset[ped][peel];
        printf("%2d:  Cut:  ",peel+1);
	for(i=0; i<c->n; i++)
	  printf("%d ",c->pid[i]);
        if(peel==npeels[ped]-1) printf("ROOT ");
        printf(" Peel:  ");
	for(i=0; i<p->n; i++)
	  printf("%d ",p->pid[i]);
        printf("\n");
      }
    }
  }
}
      
  

