 /**************************************************************

   Program:  scan
   File:     scan.c
   Author:   Zachary Skrivanek
   Date:     November 13, 2000
   Purpose:  Calculates Whittemore & Halpern's Multi-Locus 
	     Non-Parametric Statistics (ALL and PAIRS) and/or 
	     LOD scores using output from SIMPLE.  Simulates at
	     everyone point of interest vs. using reweighting.
   Modified: use cM instead of M
   Modified: 7/20/02-  1) added dumppairs (dump EIBD sharing)
	     2) simulate null using 20*maxiter.
   
************************************************************* */

#include <stdio.h> /* scanf() printf() */
#include <stdlib.h> /* pow() */
#include <math.h> /* sqrt,pow,log10 */
#include "simple.h"   /* c commands */ 
/* #include <sys/ddi.h> I don't think I need this and besides
it causes problems with allele */

#define  F 0
#define  M 1
#define  PID 0
#define  FOUNDER 0
#define  to_cmorgan      0
#define  to_recomb_frac 1
#define  MAX_NAME 25 

#define dd 0
#define dD 1
#define Dd 1
#define DD 2

#define TRUE 1
#define FALSE 0 
#define NOT_AFFECTED 1 
#define AFFECTED 2

#define sib1 0
#define sib2 1 

int main(int argc, char **argv)
{
  char   *simpleout,*score;
  FILE   *gfopen(), *runfile;

/* ReadLinkageFile stuff */
  double  **liability;
  int     no_liab_class;
  int     *nalleles;
  int     nloci;
  /* double  *distances;   */
  double  *mdistances;
 /*  int     *locusorder; */
  int     *markerorder;
  int     isdisease;
  int     nQT;
  double  disaf[2];
  double  **maf;

/* ReadPedFile */
  int    npedigrees;
  int    *pedigreesize;
  int    ***pedigree;
  aff    **affected;
  int    ****marker;
  double ***qpheno;
  char   **pednames;

/* Commands with their default values */
  cmd     c;
  char    command[MAX_NAME];

  int    nmarkers;
  int    iter;
  bool   recomb_fracs, npl=FALSE, lod=FALSE, he=FALSE;

  int    ***ancestry;
  int    *naffected;

/* FindPairs */
  prs    ***pairs;
  int    **npairs;

/* EM */
  int    ibd, ped, type, pair;  /* counter */
  int    states_sib1[2], states_sib2[2];
  double ******IBDprob;
  double ****line;
  double den;
  int    ntypes;

  double ***ppotential;
 
  double *offset; 

  double n; /* for c.maxiter.par */
  int    nweight;
  double *weight;
  double w;
  double wstart;
  double log10e;
  double **distances;
  double ***likdistances;
  double **prob;
  double ***likprob;

  int    locus;
  int    **locusorder;
  int    *numlik;

  int    i,j,k,l;
  int    ***IBDsim;
  int    ****ibdsim; /* put in to use old calc lik prog */
  int    *ibdF; /* put in to use old calc lik prog */
  int    *ibdM; /* put in to use old calc lik prog */
  int    f,m;
  int    ****IVsim;
  int    ***disIV;
  int    ***peeled;
  int    ***states,***s;
  double *null;
  double *null_squared;
  double *null_var;
  double *null_mn;
  double ***S;
  double ***p_value;
  double ***S_squared;
  double ***var; /* variance of the raw scores */
  double Score=0.0;
  double ***Z;
  double likrat,likrat_prod,**likrats;
  double ***weight_tot;
  double sum_weights;
  double **position;
  double cvsqd, cvsqd_null, cvcov;
  double **wsqr,**wcrossterm,**lik,**lod_var;
  int    tempint;
  int    **a,**bit,**num,**seen; 
  int    p1,p2;


  void   ReadLinkageFile();
  void   ReadPedFile();
  void   ParseCommands();
  void   GetAncestry();
  void   ReadRecombFracs();
  void   ReadDistances();
  void   CreateDistances_LR();
  void   CreateDistances_noLR();
  void   MapMatrix();
  void   ReadSIMPLE();
  void   SimIBDs();
  void   SimJointIBDs();
  void   GetIBDStates(); 
  double CalcScorePairs();
  double CalcScoreAll();
  double CalcLikRatio();
  void   FindPairs();
  void   EM();
  void   DumpCommands();

  void   DebugReadRun();
  void   DebugReadMarkers();
  void   DebugReadPedigree();
  void   DebugGetAncestry();
  void   DebugReadDiseasePrePPots();
  void   DebugGetAffected();
  void   DebugReadRecombFracs();
  void   DebugCreateDistances_LR();
  void   DebugCreateDistances_noLR();
  void   DebugReadDistances();
  void   DebugMapMatrix();
  void   DebugReadSIMPLE();
  void   DebugSimIBDs();
  void   DebugGetIBDStates();  
  void   DebugFindPairs();
  void   DebugIBDprob();
  
  void   PrintScores();                        
  void   DumpPairs();                        
  void   DumpProbs();                        
  void   PlantSeed();

  int    power(int, int);
  int    unbase();
  double max(double, double);


/* allocate space */
  c.load.par= (char *) calloc(MAX_NAME, sizeof(char));
  c.scan.par= (char *) calloc(MAX_NAME, sizeof(char));

/* default LociName, PedName */
  strcpy(c.load.par,"linkloci.dat");
  strcpy(c.scan.par,"ped.dat");


  if(argc>2) nerror("USAGE: scan 'commandfile (optional)' < 'simple output'" );

  log10e = log10(exp(1.0));
 
  if(argc==2){
    runfile = gfopen(argv[1],"r");
    while(fscanf(runfile,"%s",command)==1){
      if(strcmp(command,"load")==0){
        if(fscanf(runfile,"%s",c.load.par)!=1)
          nerror("expecting a file name after command 'load'");}
      else if(strcmp(command,"scan")==0){
        if(fscanf(runfile,"%s",c.scan.par)!=1)
          nerror("expecting a file name after command 'scan'");}
      else if(strcmp(command,"missingQT")==0){
        if(fscanf(runfile,"%lf",&c.missingQT.par)!=1)
          nerror("expecting a real number after command 'missingQT'");
        else c.missingQT.def=FALSE;
      }
    }
  }

  ReadLinkageFile(&nloci,
                  &mdistances,
		  &markerorder,
                  &nalleles,
                  &maf,
                  &isdisease,
                  &nQT,
                  disaf,
                  &liability,
                  &no_liab_class,
                  c.load.par);

  /* nloci = nloci - nQT; */
  nmarkers = nloci - isdisease;

  ReadPedFile(&npedigrees,
	      &pednames,
              &pedigreesize,
              &pedigree,
              &marker,
              nmarkers,
	      nQT,
	      &qpheno,
              isdisease,
              &affected,
              no_liab_class,
              c);

  ParseCommands(&c,
                argc,
                runfile,
                &nloci,
		nQT,
                nalleles,
		npedigrees,
                isdisease);

  if((strcmp(c.analysis.par,"BOTH")==0) || (strcmp(c.analysis.par,"LOD")==0)) lod=TRUE;
  if((strcmp(c.analysis.par,"BOTH")==0) || (strcmp(c.analysis.par,"NPL")==0)) npl=TRUE;
  if(c.hasemanelston.par==TRUE) he=TRUE;

  if(c.debug.par) DebugReadRun(c.debug.par,
                               c.interference.model,
                               c.interference.par,
                               nmarkers,
                               &c.file,
                               mdistances,
                               c.nsteps.par,
                               isdisease,
                               c.processorder.par,
                               c.maxiter.par,
                               c.seed.par);


  if(c.debug.par) DebugReadPedigree(npedigrees,
                                    pedigreesize,
                                    pedigree);

  if(c.debug.par) DebugReadMarker(npedigrees,
                                  pedigreesize,
                                  pedigree,
                                  nmarkers,
                                  nalleles,
                                  marker,
                                  maf,
				  nQT,
				  qpheno,
				  c.missingQT.par);


  srand48(c.seed.par);

/* this is changed in ParseCommands for SIMPLE, but in this case we are always going
to anlayze disease case if we have 'isdisease' 
This shouldn't be necessary however since ParseCommands
should produce an error if you try to do NPL anlaysis w/o
disease data 
BUT since we use the same ParseCommands in SIMLE we have to
adjust nloci now */
  nloci = nmarkers + isdisease; 
  /*c.file.nmodels = 2*nmarkers;*/ /* this is 2*nmarkers instead of nmarkers
			   since we also simulate on TOP of the
			   markers and not just in between as in
			   SIMPLE for lod scores*/

  /* c.file.nmodels = (c.nsteps.par)*(nmarkers-1)+2; */

  GetAncestry(npedigrees,
	      pedigreesize,
	      pedigree,
	      &ancestry);
  if(c.debug.par) DebugGetAncestry(npedigrees,
			     pedigreesize,
			     pedigree,
			     ancestry);
  if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
    if(c.type.par==1 || c.type.par==2) ntypes=1;
    else ntypes=2;
    FindPairs(npedigrees,
	      pedigreesize,
	      pedigree,
	      ancestry,
	      &pairs,
	      &npairs,
	      qpheno,
	      c.whichQT.par,
	      c.missingQT.par,
	      c.type.par);
  
    if(c.debug.par) DebugFindPairs(npedigrees,
				   npairs,
				   pairs,
				   c.type.par);
  }

/* If units are not specified then we determine if we have 
 cM or recomb fracs using the preplink
 criteria that if at least one > .5 we have cM */

  if(c.units.def){
    recomb_fracs = TRUE;
    strcpy(c.units.par,"rec-frac");
    for(i=0; i < nmarkers-1; i++)
      if(mdistances[i] > 0.5) {
	recomb_fracs = FALSE;
        strcpy(c.units.par,"cM");
      }
  }
  else{
    if(strcmp(c.units.par,"cM")==0) recomb_fracs=FALSE;
    else if(strcmp(c.units.par,"rec-frac")==0) recomb_fracs=TRUE;
    else fprintf(stderr,"'units %s' is not a valid command!\n",c.units.par); 
    nerror("Command file error!");
  }


  if(!c.file.def) /* the default is that there is no file and we create distances */
    if(recomb_fracs==TRUE){
      ReadRecombFracs(nloci,
                      c.file.nmodels,
                      &locusorder,
                      &distances, /* actually recomb fracs */
                      &numlik,
                      &likdistances);
      if(c.debug.par) DebugReadRecombFracs(nloci,
                                           c.file.nmodels,
                                           locusorder,
                                           distances,
                                           numlik,
                                           likdistances);
    }
    else{
      ReadDistances(nloci,
                    &locusorder,
                    &distances, 
                    &numlik,
                    &likdistances,
		    c.file);
      if(c.debug.par) DebugReadDistances(nmarkers,
                                         c.file.nmodels,
                                         locusorder,
                                         distances,
                                         numlik,
                                         likdistances);
    } 
  else {
    if(recomb_fracs==TRUE)
      MapVector(nmarkers-1, mdistances, to_cmorgan);
    if(c.reweight.par){
      CreateDistances_LR(nmarkers, /* this is different than in simple */
                         &c.file.nmodels,
                         c.nsteps.par,
                         mdistances,
                         &locusorder,
                         &distances,
                         &numlik,
                         &likdistances);
         if(c.debug.par) DebugCreateDistances_LR(nmarkers,
                                                 c.file.nmodels,
                                                 locusorder,
                                                 distances,
                                                 numlik,
                                                 likdistances);
    }
    else{
      CreateDistances_noLR(nmarkers, /* this is different than in simple */
                           &c.file.nmodels,
                           c.nsteps.par,
                           mdistances,
                           &locusorder,
                           &distances,
                           &numlik,
                           &likdistances);
           if(c.debug.par) DebugCreateDistances_noLR(nmarkers,
                                                     c.file.nmodels,
                                                     locusorder,
                                                     distances,
                                                     numlik,
                                                     likdistances);
    }
  } 
  switch(c.interference.model){
  case 0:
    /* map cmorgans to recomb fracs */
      MapMatrix(nloci,
                c.file.nmodels,
                distances,
                numlik,
                likdistances,
                isdisease,
                to_recomb_frac);
      if(c.debug.par) DebugMapMatrix(nloci,
                               c.file.nmodels,
                               distances,
                               numlik,
                               likdistances,
                               isdisease,
                               to_recomb_frac);
      Calcibdprobs(nloci,
                   c.file.nmodels,
                   locusorder,
                   distances,  /* actually transformed to rec fracs */
                   &prob,
                   numlik,
                   likdistances, /* actually transformed to rec fracs */
                   &likprob);

    break;      
    case 19:              
      CHS(nloci,
          c.file.nmodels,
          1,  /* isdisease is TRUE regardless of SIMPLE */
          locusorder,
          distances,
          numlik,
          likdistances,
          &prob,
          &likprob,
          c.interference.par);
    break;
    default:
      nerror(stderr,"There is no proper interference model.");
    break;
  }

  Calcmarginprobs(nloci,
                  c.file.nmodels,
                  prob,
                  numlik,
                  likprob);
  if(c.debug.par) DebugCalcmarginprobs(nloci,
                                 c.file.nmodels,
                                 prob,
                                 numlik,
                                 likprob);

  if(c.debug.par) DebugProbtable(nloci,
                                 c.file.nmodels,
                                 prob,
                                 numlik,
                                 likprob);

  /*******************
  ** allocate space **
  *******************/

  if(lod){
    wsqr = (double **) calloc(nmarkers+(nmarkers==1), sizeof(double *));
    wcrossterm = (double **) calloc(nmarkers+(nmarkers==1)-1, sizeof(double *));
    lod_var = (double **) calloc(nmarkers+(nmarkers==1), sizeof(double *));
    lik = (double **) calloc(nmarkers+(nmarkers==1), sizeof(double *));
    likrats = (double **) calloc(nmarkers+(nmarkers==1), sizeof(double *));
  }
    
  IBDsim = (int ***) calloc(npedigrees, sizeof(int **));
  ibdsim = (int ****) calloc(npedigrees, sizeof(int ***));
  null_mn = (double *) calloc(npedigrees, sizeof(double));
  null_var = (double *) calloc(npedigrees, sizeof(double));
  weight_tot = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
  disIV = (int ***) calloc(npedigrees, sizeof(int **));
  states = (int ***) calloc(npedigrees, sizeof(int **));

  if(npl){
    S = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
    p_value = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
    S_squared = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
    var = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
    Z = (double ***) calloc(c.file.nmodels-1, sizeof(double **));
    null = (double *) calloc(npedigrees, sizeof(double));
    null_squared = (double *) calloc(npedigrees, sizeof(double));
  }

  for(i=0; i<npedigrees; i++){
    IBDsim[i] = (int **) calloc(pedigreesize[i], sizeof(int *));
    ibdsim[i] = (int ***) calloc(pedigreesize[i], sizeof(int **));
    disIV[i] = (int **) calloc(pedigreesize[i], sizeof(int *));
    states[i] = (int **) calloc(pedigreesize[i], sizeof(int *));
    for(j=0; j<pedigreesize[i]; j++){
      IBDsim[i][j] = (int *) calloc(2, sizeof(int));
      ibdsim[i][j] = (int **) calloc(nloci, sizeof(int *));
      for(locus=0; locus<nloci; locus++)
        ibdsim[i][j][locus] = (int *) calloc(2, sizeof(int));
      disIV[i][j] = (int *) calloc(2, sizeof(int));
      states[i][j] = (int *) calloc(2, sizeof(int));
    }
  }
  
  ibdF = (int *) calloc(nloci, sizeof(int));
  ibdM = (int *) calloc(nloci, sizeof(int));
 
  if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
    IBDprob = (double ******) calloc(c.file.nmodels-1, sizeof(double *****));
    line = (double ****) calloc(c.file.nmodels-1, sizeof(double ***));
  }

  for(i=0; i < c.file.nmodels-1; i++){
    weight_tot[i] = (double **) calloc(numlik[i+1], sizeof(double *));
    if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
      IBDprob[i] = (double *****) calloc(numlik[i+1], sizeof(double ****));
      line[i] = (double ***) calloc(numlik[i+1], sizeof(double **));
    }
    if(npl){
      S[i] = (double **) calloc(numlik[i+1], sizeof(double *));
      p_value[i] = (double **) calloc(numlik[i+1], sizeof(double *));
      S_squared[i] = (double **) calloc(numlik[i+1], sizeof(double *));
      var[i] = (double **) calloc(numlik[i+1], sizeof(double *));
      Z[i] = (double **) calloc(numlik[i+1], sizeof(double *));
    }
    for(j=0; j<numlik[i+1]; j++){
      if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
        IBDprob[i][j] = (double ****) calloc(ntypes, sizeof(double ***));
	/* line[i][j] = (double **) calloc(ntypes, sizeof(double *)); 7/19/02 */
	line[i][j] = (double **) calloc(1, sizeof(double *)); 
	line[i][j][0] = (double *) calloc(3, sizeof(double)); /* 7/19/02 */
	for(type=0; type<ntypes; type++){ 
	  /* line[i][j][type] = (double *) calloc(3, sizeof(double)); 7/19/02 */
          IBDprob[i][j][type] = (double ***) calloc(npedigrees, sizeof(double **));
          for(ped=0; ped<npedigrees; ped++){
            IBDprob[i][j][type][ped] = (double **) calloc(npairs[type][ped], sizeof(double *));
	    for(pair=0; pair<npairs[type][ped]; pair++)
              IBDprob[i][j][type][ped][pair] = (double *) calloc(3, sizeof(double));
          } 
        } 
      } 
      weight_tot[i][j] = (double *) calloc(npedigrees, sizeof(double));
      if(npl){
        S[i][j] = (double *) calloc(npedigrees, sizeof(double));
        p_value[i][j] = (double *) calloc(npedigrees, sizeof(double));
        S_squared[i][j] = (double *) calloc(npedigrees, sizeof(double));
        var[i][j] = (double *) calloc(npedigrees, sizeof(double));
        Z[i][j] = (double *) calloc(npedigrees, sizeof(double));
      }
    }
  }


  if(lod){
    nweight=nmarkers+1+nmarkers-1; /* at markers, null case and
     in between markers */
    if(nmarkers==1) nweight=3; /* 6/11/02 */
    wsqr[0] = (double *) calloc(numlik[0], sizeof(double));
    lod_var[0] = (double *) calloc(numlik[0], sizeof(double));
    lik[0] = (double *) calloc(numlik[0], sizeof(double));
    likrats[0] = (double *) calloc(1, sizeof(double));
    if(nmarkers==1){
      wsqr[1] = (double *) calloc(numlik[0], sizeof(double));
      lod_var[1] = (double *) calloc(numlik[0], sizeof(double));
      lik[1] = (double *) calloc(numlik[0], sizeof(double));
      likrats[1] = (double *) calloc(1, sizeof(double));
      wcrossterm[0] = (double *) calloc(1, sizeof(double));
    }
    for(i=1; i<nmarkers; i++){
    /*
      wsqr[i] = (double *) calloc(numlik[i+nmarkers], sizeof(double));
      wcrossterm[i] = (double *) calloc(numlik[i+nmarkers], sizeof(double));
      lod_var[i] = (double *) calloc(numlik[i+nmarkers], sizeof(double));
      lik[i] = (double *) calloc(numlik[i+nmarkers], sizeof(double));
      likrats[i] = (double *) calloc(numlik[i+nmarkers], sizeof(double));
      */
      wsqr[i] = (double *) calloc(c.nsteps.par+1, sizeof(double));
      wcrossterm[i] = (double *) calloc(c.nsteps.par+1, sizeof(double));
      lod_var[i] = (double *) calloc(c.nsteps.par+1, sizeof(double));
      lik[i] = (double *) calloc(c.nsteps.par+1,sizeof(double));
      likrats[i] = (double *) calloc(c.nsteps.par+1, sizeof(double));
    }
  } 
  else nweight=nmarkers; /* at markers */

  weight = (double *) calloc(nweight,sizeof(double));

  peeled = (int ***) calloc(npedigrees, sizeof(int **));
  for(i=0; i<npedigrees; i++)  {
    peeled[i] = (int **) calloc(pedigreesize[i], sizeof(int *));
    for(j=0; j<pedigreesize[i]; j++)
      peeled[i][j] = (int *) calloc(2, sizeof(int));
  }

  naffected = (int *) calloc(npedigrees, sizeof(int));
  for(i=0; i<npedigrees; i++)
    for(j=0; j<pedigreesize[i]; j++)
      if(affected[i][j].status==AFFECTED)
        ++naffected[i];
  
  a = (int **) calloc(npedigrees, sizeof(int *));
  bit = (int **) calloc(npedigrees, sizeof(int *));
  seen = (int **) calloc(npedigrees, sizeof(int *));
  num = (int **) calloc(npedigrees, sizeof(int *));
  s = (int ***) calloc(npedigrees, sizeof(int **));
  for(i=0; i<npedigrees; i++){
    a[i] = (int *) calloc(naffected[i], sizeof(int));
    bit[i] = (int *) calloc(naffected[i], sizeof(int));
    seen[i] = (int *) calloc(naffected[i], sizeof(int));
    num[i] = (int *) calloc(naffected[i], sizeof(int));
    s[i] = (int **) calloc(naffected[i], sizeof(int *));
    for(j=0; j<naffected[i]; j++)
      s[i][j] = (int *) calloc(2, sizeof(int));
  }


  position = (double **) calloc(c.file.nmodels,sizeof(double *));
  for(i=0; i<c.file.nmodels; i++)
    position[i] = (double *) calloc(numlik[i],sizeof(double));

  if(isdisease){
    for(i=1; i<c.file.nmodels; i++)
    {
      for(k=0; k<nloci; k++)
        if(locusorder[i][k] == nloci) break;
      for(j=0; j<numlik[i]; j++)
        for(l=0; l<k; l++)
          position[i][j] += likdistances[i][j][l];
    }
  }


  /*********************
  ** read in IBD data **
  *********************/
  scanf("%d",&tempint);
  for(i=0; i<nweight; i++)
    scanf("%le",&weight[i]);
  rewind(stdin);

  /* for numerical stability we divide each weight by their sum */
  /* offset = 0.0;
  for(i=0; i< nmarkers; i++)
    offset -= log10(weight[i]);
  wstart = pow(10.0,offset); 
*/
  offset = (double *) calloc(nweight, sizeof(double));
  for(i=0; i< nweight; i++)
    offset[i] = weight[i];

  sum_weights=0.0;
  for(iter=0; iter<c.maxiter.par; iter++){

    ReadSIMPLE(IBDsim,
               weight,
	       nweight,
	       likrats,
	       nmarkers,
	       numlik,
               ancestry,
               pedigreesize,
               npedigrees,
	       c,
	       iter);
    if(c.debug.par) DebugReadSIMPLE(IBDsim,
                                    weight,
			            nweight,
	                            likrats,
	                            nmarkers,
	                            numlik,
                                    ancestry,
                                    pedigree,
                                    pedigreesize,
                                    npedigrees,
			            c);
  
    /* w=wstart;  */
    /* this was a mistake if(npl) 
      for(i=0; i<nmarkers; i++)
        w *= weight[i];
    else
      for(i=0; i<2*nmarkers; i++)
        w *= weight[i]; */

/*     for(i=0; i<nmarkers; i++)
      w *= weight[i]; */


    w=1.0;
    if(c.dumpscores.par) {
      for(i=0; i<nmarkers; i++)
        w *= weight[i];
      printf("%d %le ",iter+1,w);
    }

    w=1.0;
    for(i=0; i<nmarkers; i++)
      w *= weight[i]/offset[i];

    sum_weights += w;
if(c.dumplikrats.par) fprintf(stderr,"%d %le ",iter+1,w);
    /* Get the likelihoods if we are to analyze LOD scores */
    if(lod)
    {
      lik[0][0] += w * weight[nloci+0-1]; 
      wsqr[0][0] += pow(w * weight[nloci+0-1],2.0);
      for(i=1; i<nmarkers; i++)
	for(j=0; j<c.nsteps.par+1; j++){
          lik[i][j] += w * weight[i+nloci-1] * likrats[i][j]; 
          wsqr[i][j] += pow(w * weight[i+nloci-1] * likrats[i][j],2.0);
          wcrossterm[i][j] += w * weight[nloci+0-1] * w * weight[i+nloci-1] * likrats[i][j]; 
        }
      if(nmarkers==1){
          lik[1][0] += w * weight[1+nloci-1] ; 
          wsqr[1][0] += pow(w * weight[1+nloci-1],2.0);
          wcrossterm[0][0] += w * weight[nloci+0-1] * w * weight[1+nloci-1]; 
      }
    } 

  if(npl || he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE)
    for(i=1; i<c.file.nmodels; i++){
    
        /* simulate the IBD info at the disease locations */
	if(c.joint.par)
          SimJointIBDs(IBDsim,
	               disIV,
                       &prob[i][0],
	               ancestry,
                       pedigreesize,
                       npedigrees,
                       nloci);
        else
          SimIBDs(IBDsim,
	          disIV,
                  &prob[i][0],
	          ancestry,
                  pedigreesize,
                  npedigrees,
                  nloci);
        if(c.debug.par) DebugSimIBDs(IBDsim,
			       disIV,
                               ancestry,
                               pedigree,
                               pedigreesize,
                               npedigrees,
			       &locusorder[i][0],
                               nloci); 

        for(j=0; j<npedigrees; j++)  
          for(k=0; k<pedigreesize[j]; k++)
	    for(l=0; l<2; l++)
              peeled[j][k][l] = FALSE; 
  
        GetIBDStates(states,
                     disIV,
                     ancestry,
                     npedigrees,
                     pedigreesize,
		     peeled);

        if(c.debug.par) DebugGetIBDStates(states,
                                          pedigree,
                                          ancestry,
                                          pedigreesize,
                                          npedigrees);
        if(npl){ 
          for(j=0; j<npedigrees; j++){
            l=0;
            for(k=0; k<pedigreesize[j]; k++)
	      if(affected[j][k].status==AFFECTED){
                s[j][l][F]=states[j][k][F];
                s[j][l][M]=states[j][k][M];
                ++l;
              }
          }
        }

        /* for(ped=0; ped<npedigrees; ped++) to use Mark's calc lik 
          for(j=0; j<pedigreesize[ped]; j++) 
            if(!ancestry[ped][j][FOUNDER]){
              f = IBDsim[ped][j][F]+disIV[ped][j][F]*power(3,nmarkers);
              m = IBDsim[ped][j][M]+disIV[ped][j][M]*power(3,nmarkers);
              basen(nloci, 3, f, ibdF);
              basen(nloci, 3, m, ibdM);
              for(locus=0; locus < nloci; locus++){
                ibdsim[ped][i][locus][F]=ibdF[locus];  
                ibdsim[ped][i][locus][M]=ibdM[locus];  
              }
           }  */


        for(ped=0; ped<npedigrees; ped++){
	  if(npl){
            if(strcmp(c.score.par,"PAIRS")==0) 
	      Score= CalcScorePairs(&s[ped][0],
          	                    naffected[ped]);
  
            else Score = CalcScoreAll(&s[ped][0],
                                      naffected[ped],
	            	              &a[ped][0],
			              &bit[ped][0],
			              &seen[ped][0],
			              &num[ped][0]); 
          } /* if npl */
	  /* do the reweighting to get other interior points */
	  if(c.dumplikrats.par) fprintf(stderr,"%le ",Score);
	  if(c.dumpscores.par) printf("%le ",Score);
          for(j=0; j<numlik[i]; j++){
            /* likrat = CalcLikRatio(npedigrees,
                                    pedigreesize,
                                    ancestry,
                                    ibdsim,
                                    ibdF,
                                    ibdM,
                                    nloci,
                                    i,
                                    j,
                                    prob,
                                    likprob); */

            likrat = CalcLikRatio(&ancestry[ped][0],
			            pedigreesize[ped],
				    &affected[ped][0],
                                    &IBDsim[ped][0],
                                    &disIV[ped][0], 
		                    &prob[i][0],
                                    &likprob[i][j][0],
                                    nmarkers);

            if(npl){ 
	      if(c.reweight.par){
                S[i-1][j][ped] += Score * likrat * w; 
                S_squared[i-1][j][ped] += pow(Score,2.0) * likrat * w;
	        if(c.dumplikrats.par) fprintf(stderr,"%le ",likrat);
	        if(c.dumpscores.par) printf("%le ",likrat);
              }
	      else{
                S[i-1][j][ped] += Score *  w;
                S_squared[i-1][j][ped] += pow(Score, 2.0)*w;
              }
	      if(c.wstar.par)
                weight_tot[i-1][j][ped] += likrat * w; 
	      else
                weight_tot[i-1][j][ped] += w;
           /* S_squared[i-1][j][ped] += pow(Score * likrat * w, 2.0); */
            } 
            if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
	      for(type=0; type<ntypes; type++)
	        for(pair=0; pair<npairs[type][ped]; pair++){
		  states_sib1[F]=states[ped][pairs[type][ped][pair].index[sib1]][F];
	          states_sib1[M]=states[ped][pairs[type][ped][pair].index[sib1]][M];
	          states_sib2[F]=states[ped][pairs[type][ped][pair].index[sib2]][F];
	          states_sib2[M]=states[ped][pairs[type][ped][pair].index[sib2]][M];
                  ibd=0;
	          for(p1=0; p1<2; p1++)
	            for(p2=0; p2<2; p2++)
		      if(states_sib1[p1]==states_sib2[p2]) ++ibd;
	          if(c.reweight.par)
	            IBDprob[i-1][j][type][ped][pair][ibd] += likrat * w;
                  else
	            IBDprob[i-1][j][type][ped][pair][ibd] += w;
              } /* pair */  
	    } /* if he */
          } /* j numlik */
        } /* ped */
    } /* nmodels */
    if(c.dumplikrats.par) fprintf(stderr,"\n"); 
    if(c.dumpscores.par) printf("\n"); 
  } /* iter */
  fclose(stdin);

  n = (double) c.maxiter.par;
  if(npl)
    for(i=1; i< c.file.nmodels; i++)
      for(j=0; j<numlik[i]; j++)
        for(k=0; k<npedigrees; k++){
	  if(c.wstar.par){
	    if(weight_tot[i-1][j][k] != 0.0){ 
              S[i-1][j][k] /= weight_tot[i-1][j][k];/* offset cancels out */
              S_squared[i-1][j][k] /= weight_tot[i-1][j][k];
            }
          }
          else
	    if(sum_weights != 0.0){ 
              S[i-1][j][k] /= sum_weights; /* offset cancels out */
              S_squared[i-1][j][k] /= sum_weights; 
            }
	  var[i-1][j][k] = S_squared[i-1][j][k] - pow(S[i-1][j][k],2.0);
	  /* var[i-1][j][k] *= power(n,2)/(n-1); */
          var[i-1][j][k] = max(var[i-1][j][k], 0.0); /* avoid rounding error */
        }

  /* free(weight_tot);
  free(S_squared);*/

  
  if(lod){
    /* calculate the var of the LOD score */
    /* calculate cv^2 at in null case d=infinity */
    /* lik[0][0] /=  (double) c.maxiter.par; */
    cvsqd_null = (wsqr[0][0]/pow(lik[0][0],2.0) * (double)
    power(c.maxiter.par,2) - (double) c.maxiter.par)
                   / ((double) (c.maxiter.par-1));
    /* C likes to give negative values to 0 */
    if(cvsqd_null < 0.0) cvsqd_null=0.0;
    /* redo for nmarkers==1 some day */
    for(i=1; i<nmarkers; i++)
      for(j=0; j<c.nsteps.par+1; j++)
        if(lik[i][j] == 0.0)
          lod_var[i][j] = 0.0;
        else{
          /* lik[i][j] /=  (double) c.maxiter.par; */
        /* calculate cvsqd at location on chromosome */
          cvsqd = (wsqr[i][j]/ pow(lik[i][j],2.0) * (double)
	  power(c.maxiter.par,2) - (double) c.maxiter.par)
                     / ((double) (c.maxiter.par-1));
       /* calculate cov */
          if(cvsqd< 0.0) cvsqd=0.0;
          cvcov = wcrossterm[i][j]/(lik[i][j] * lik[0][0])*(double)
	  power(c.maxiter.par,2) - (double) c.maxiter.par;
	  cvcov /= ((double) c.maxiter.par - 1);
          if(cvcov < 0.0) cvcov=0.0;
          lod_var[i][j] =  cvsqd + cvsqd_null - 2*cvcov;
	  lod_var[i][j] /= (double) power(c.maxiter.par,2);
	}
  }

  if(he || c.dumppairs.par==TRUE || c.dumpprobs.par==TRUE){
    for(iter=0; iter < c.maxiter.par; iter++)
      for(i=1; i< c.file.nmodels; i++)
        for(j=0; j<numlik[i]; j++)
          for(ped=0; ped<npedigrees; ped++)
	    for(type=0; type<ntypes; type++)
	      for(pair=0; pair<npairs[type][ped]; pair++){
	        den=0.0;
	        for(ibd=0; ibd<3; ibd++)
		  den += IBDprob[i-1][j][type][ped][pair][ibd];
	        for(ibd=0; ibd<3; ibd++)
		  IBDprob[i-1][j][type][ped][pair][ibd] /= den;
              }

    if(c.debug.par) DebugIBDprob(c.file.nmodels,
				 numlik,
				 npedigrees,
				 pedigreesize,
				 npairs,
				 position,
				 nmarkers,
				 IBDprob,
				 ntypes);
  } 
  /* EM algorithm */
  if(he)
    EM(npedigrees,
       pairs,
       npairs,
       c.file.nmodels,
       numlik,
       IBDprob,
       line);
    

  if(npl)
  {
    for(iter=0; iter < (20*c.maxiter.par); iter++){
      /* Simulate null distribution. 
      The null distribution is simulated by disIV
      in this case since for model 0 the disease is unlinked */
      SimIBDs(IBDsim,
	      disIV,
              &prob[0][0],
	      ancestry,
              pedigreesize,
              npedigrees,
              nloci);
      if(c.debug.par) DebugSimIBDs(IBDsim,
			     disIV,
                             ancestry,
                             pedigree,
                             pedigreesize,
                             npedigrees,
                             nloci);
  
      for(i=0; i<npedigrees; i++)  
        for(j=0; j<pedigreesize[i]; j++)
	  for(k=0; k<2; k++)
            peeled[i][j][k] = 0; 
  
      GetIBDStates(states,
                   disIV,
                   ancestry,
                   npedigrees,
                   pedigreesize,
		   peeled);
      if(c.debug.par) DebugGetIBDStates(states,
                                  pedigree,
                                  ancestry,
                                  pedigreesize,
                                  npedigrees);
  
      for(j=0; j<npedigrees; j++){
        l=0;
        for(k=0; k<pedigreesize[j]; k++)
	  if(affected[j][k].status==AFFECTED){
            s[j][l][F]=states[j][k][F];
            s[j][l][M]=states[j][k][M];
            ++l;
          }
      } 
  
      for(k=0; k<npedigrees; k++){ /* calculate null distribution */
        if(strcmp(c.score.par,"PAIRS")==0) Score= CalcScorePairs(&s[k][0],
                                        naffected[k]);
  
        else Score = CalcScoreAll(&s[k][0],
                                  naffected[k],
	         	          &a[k][0],
				  &bit[k][0],
				  &seen[k][0],
				  &num[k][0]); 
  
        null[k] += Score;
        null_squared[k] += pow(Score,2.0);
        for(i=1; i<c.file.nmodels; i++)
          for(j=0; j<numlik[i]; j++)
            if(Score >= S[i-1][j][k]) ++p_value[i-1][j][k];
  
      }
    } /* iter */

    for(i=0; i<npedigrees; i++) {
      null_mn[i] = null[i] / (double) (20*c.maxiter.par);
      null_var[i] = null_squared[i]-pow(null_mn[i],2.0) * ((double) (20*c.maxiter.par)); 
      null_var[i] /= (double) (20*c.maxiter.par - 1); 
    }
    
    /* standardize the scores and average the p-values */
    for(i=1; i< c.file.nmodels; i++)
      for(j=0; j<numlik[i]; j++)
        for(k=0; k<npedigrees; k++){
          Z[i-1][j][k] = (S[i-1][j][k] - null_mn[k])/sqrt(null_var[k]);
          p_value[i-1][j][k] /= (double) (20*c.maxiter.par);
        }
  } /* if calculate NPL */

/* This is a hack to just print out the Z scores and sd of the raw scores for simulation purposes */
/*
  for(i=1; i<c.file.nmodels; i++)
    for(j=0; j<numlik[i]; j++)
      for(k=0; k<npedigrees; k++)
          printf("%9.6lf ", Z[i-1][j][k]);
  printf("\n");
  for(i=1; i<c.file.nmodels; i++)
    for(j=0; j<numlik[i]; j++)
      for(k=0; k<npedigrees; k++)
          printf("%9.6lf ", sqrt(var[i-1][j][k]));
  printf("\n");
*/

  PrintScores(S,
	      var,
              p_value, 
              Z,
              null_mn,
              null_var,
              lik,
	      lod_var,
	      line,
	      IBDprob,
	      position,
              nmarkers,
              c,
              numlik,
              npedigrees,
	      pairs,
	      npairs); 

  if(c.dumppairs.par==TRUE)
    DumpPairs(IBDprob,
              npedigrees,
              &npairs[ntypes-1][0],
              position,
	      numlik,
	      pairs,
	      c);

  if(c.dumpprobs.par==TRUE)
    DumpProbs(IBDprob,
              npedigrees,
              &npairs[ntypes-1][0],
              position,
	      numlik,
	      pairs,
	      c);

  if(!c.sim.par)
    DumpCommands(c,
                 argv,
                 argc,
		 npedigrees,
                 nloci,
		 numlik,
		 c);

  PlantSeed();
  exit(0);

}

